@@ -12,7 +12,10 @@ import socket
 import sys
 import threading
 import time
+import urlparse
+from Queue import Empty
 from StringIO import StringIO
+from abc import ABCMeta, abstractmethod
 from collections import defaultdict, OrderedDict
 from multiprocessing import Queue
 
@@ -97,6 +100,7 @@ class TestEnvironment(object):
         self.test_path = test_path
         self.server = None
         self.config = None
+        self.external_config = None
         self.test_server_port = options.pop("test_server_port", True)
         self.options = options if options is not None else {}
         self.required_files = options.pop("required_files", [])
@@ -105,11 +109,11 @@ class TestEnvironment(object):
     def __enter__(self):
         self.copy_required_files()
 
-        config = self.load_config()
-        serve.set_computed_defaults(config)
+        self.config = self.load_config()
+        serve.set_computed_defaults(self.config)
 
         serve.logger = serve.default_logger("info")
-        self.config, self.servers = serve.start(config)
+        self.external_config, self.servers = serve.start(self.config)
         return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
@@ -369,14 +373,38 @@ class TestFilter(object):
 
 
 class TestLoader(object):
-    def __init__(self, tests_root, metadata_root, test_filter, run_info):
+    def __init__(self, tests_root, metadata_root, test_types, test_filter, run_info,
+                 chunk_type="none", total_chunks=1, chunk_number=1):
         self.tests_root = tests_root
         self.metadata_root = metadata_root
+        self.test_types = test_types
         self.test_filter = test_filter
         self.run_info = run_info
         self.manifest_path = os.path.join(self.metadata_root, "MANIFEST.json")
         self.manifest = self.load_manifest()
         self.tests = None
+        self.disabled_tests = None
+
+        self.chunk_type = chunk_type
+        self.total_chunks = total_chunks
+        self.chunk_number = chunk_number
+
+        self.chunker = {"none": Unchunked,
+                        "hash": HashChunker,
+                        "equal_time": EqualTimeChunker}[chunk_type](total_chunks,
+                                                                    chunk_number)
+
+        self._test_ids = None
+        self._load_tests()
+
+    @property
+    def test_ids(self):
+        if self._test_ids is None:
+            self._test_ids = []
+            for test_dict in [self.disabled_tests, self.tests]:
+                for test_type in self.test_types:
+                    self._test_ids += [item.id for item in test_dict[test_type]]
+        return self._test_ids
 
     def create_manifest(self):
         logger.info("Creating test manifest")
@@ -400,11 +428,11 @@ class TestLoader(object):
     def load_expected_manifest(self, test_path):
         return manifestexpected.get_manifest(self.metadata_root, test_path, self.run_info)
 
-    def iter_tests(self, test_types, chunker=None):
-        manifest_items = self.test_filter(self.manifest.itertypes(*test_types))
+    def iter_tests(self):
+        manifest_items = self.test_filter(self.manifest.itertypes(*self.test_types))
 
-        if chunker is not None:
-            manifest_items = chunker(manifest_items)
+        if self.chunker is not None:
+            manifest_items = self.chunker(manifest_items)
 
         for test_path, tests in manifest_items:
             expected_file = self.load_expected_manifest(test_path)
@@ -413,34 +441,19 @@ class TestLoader(object):
                 test_type = manifest_test.item_type
                 yield test_path, test_type, test
 
-    def get_disabled(self, test_types):
-        rv = defaultdict(list)
-
-        for test_path, test_type, test in self.iter_tests(test_types):
-            if test.disabled():
-                rv[test_type].append(test)
-
-        return rv
-
-    def load_tests(self, test_types, chunk_type, total_chunks, chunk_number):
+    def _load_tests(self):
         """Read in the tests from the manifest file and add them to a queue"""
-        rv = defaultdict(list)
+        tests = {"enabled":defaultdict(list),
+                 "disabled":defaultdict(list)}
 
-        chunker = {"none": Unchunked,
-                   "hash": HashChunker,
-                   "equal_time": EqualTimeChunker}[chunk_type](total_chunks,
-                                                               chunk_number)
+        for test_path, test_type, test in self.iter_tests():
+            key = "enabled" if not test.disabled() else "disabled"
+            tests[key][test_type].append(test)
 
-        for test_path, test_type, test in self.iter_tests(test_types, chunker):
-            if not test.disabled():
-                rv[test_type].append(test)
-
-        return rv
-
-    def get_groups(self, test_types, chunk_type="none", total_chunks=1, chunk_number=1):
-        if self.tests is None:
-            self.tests = self.load_tests(test_types, chunk_type, total_chunks, chunk_number)
+        self.tests = tests["enabled"]
+        self.disabled_tests = tests["disabled"]
 
+    def groups(self, test_types, chunk_type="none", total_chunks=1, chunk_number=1):
         groups = set()
 
         for test_type in test_types:
@@ -450,20 +463,82 @@ class TestLoader(object):
 
         return groups
 
-    def queue_tests(self, test_types, chunk_type, total_chunks, chunk_number):
-        if self.tests is None:
-            self.tests = self.load_tests(test_types, chunk_type, total_chunks, chunk_number)
 
-        tests_queue = defaultdict(Queue)
-        test_ids = []
+class TestSource(object):
+    __metaclass__ = ABCMeta
 
-        for test_type in test_types:
-            for test in self.tests[test_type]:
-                tests_queue[test_type].put(test)
-                test_ids.append(test.id)
+    @abstractmethod
+    def queue_tests(self, test_queue):
+        pass
 
-        return test_ids, tests_queue
+    @abstractmethod
+    def requeue_test(self, test):
+        pass
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *args, **kwargs):
+        pass
+
+
+class SingleTestSource(TestSource):
+    def __init__(self, test_queue):
+        self.test_queue = test_queue
+
+    @classmethod
+    def queue_tests(cls, test_queue, test_type, tests):
+        for test in tests[test_type]:
+            test_queue.put(test)
+
+    def get_queue(self):
+        if self.test_queue.empty():
+            return None
+        return self.test_queue
+
+    def requeue_test(self, test):
+        self.test_queue.put(test)
+
+class PathGroupedSource(TestSource):
+    def __init__(self, test_queue):
+        self.test_queue = test_queue
+        self.current_queue = None
+
+    @classmethod
+    def queue_tests(cls, test_queue, test_type, tests, depth=None):
+        if depth is True:
+            depth = None
+
+        prev_path = None
+        group = None
+
+        for test in tests[test_type]:
+            path = urlparse.urlsplit(test.url).path.split("/")[1:-1][:depth]
+            if path != prev_path:
+                group = []
+                test_queue.put(group)
+                prev_path = path
+
+            group.append(test)
+
+    def get_queue(self):
+        if not self.current_queue or self.current_queue.empty():
+            try:
+                data = self.test_queue.get(block=True, timeout=1)
+                self.current_queue = Queue()
+                for item in data:
+                    self.current_queue.put(item)
+            except Empty:
+                return None
+
+        return self.current_queue
+
+    def requeue_test(self, test):
+        self.current_queue.put(test)
+
+    def __exit__(self, *args, **kwargs):
+        if self.current_queue:
+            self.current_queue.close()
 
 class LogThread(threading.Thread):
     def __init__(self, queue, logger, level):
@@ -516,18 +591,20 @@ def list_test_groups(tests_root, metadata_root, test_types, product, **kwargs):
     run_info = wpttest.get_run_info(metadata_root, product, debug=False)
     test_filter = TestFilter(include=kwargs["include"], exclude=kwargs["exclude"],
                              manifest_path=kwargs["include_manifest"])
-    test_loader = TestLoader(tests_root, metadata_root, test_filter, run_info)
+    test_loader = TestLoader(tests_root, metadata_root, test_types, test_filter, run_info)
 
-    for item in sorted(test_loader.get_groups(test_types)):
+    for item in sorted(test_loader.groups()):
         print item
 
 
 def list_disabled(tests_root, metadata_root, test_types, product, **kwargs):
+    do_test_relative_imports(tests_root)
+
     rv = []
     run_info = wpttest.get_run_info(metadata_root, product, debug=False)
-    test_loader = TestLoader(tests_root, metadata_root, TestFilter(), run_info)
+    test_loader = TestLoader(tests_root, metadata_root, test_types, TestFilter(), run_info)
 
-    for test_type, tests in test_loader.get_disabled(test_types).iteritems():
+    for test_type, tests in test_loader.disabled_tests.iteritems():
         for test in tests:
             rv.append({"test": test.id, "reason": test.disabled()})
     print json.dumps(rv, indent=2)
@@ -565,9 +642,25 @@ def run_tests(config, tests_root, metadata_root, product, **kwargs):
         if "test_loader" in kwargs:
             test_loader = kwargs["test_loader"]
         else:
-            test_filter = TestFilter(include=kwargs["include"], exclude=kwargs["exclude"],
+            test_filter = TestFilter(include=kwargs["include"],
+                                     exclude=kwargs["exclude"],
                                      manifest_path=kwargs["include_manifest"])
-            test_loader = TestLoader(tests_root, metadata_root, test_filter, run_info)
+            test_loader = TestLoader(tests_root,
+                                     metadata_root,
+                                     kwargs["test_types"],
+                                     test_filter,
+                                     run_info,
+                                     kwargs["chunk_type"],
+                                     kwargs["total_chunks"],
+                                     kwargs["this_chunk"])
+
+        if kwargs["run_by_dir"] is False:
+            test_source_cls = SingleTestSource
+            test_source_kwargs = {}
+        else:
+            # A value of None indicates infinite depth
+            test_source_cls = PathGroupedSource
+            test_source_kwargs = {"depth": kwargs["run_by_dir"]}
 
         logger.info("Using %i client processes" % kwargs["processes"])
 
@@ -578,22 +671,22 @@ def run_tests(config, tests_root, metadata_root, product, **kwargs):
                 logger.critical("Error starting test environment: %s" % e.message)
                 raise
 
-            base_server = "http://%s:%i" % (test_environment.config["host"],
-                                            test_environment.config["ports"]["http"][0])
+            base_server = "http://%s:%i" % (test_environment.external_config["host"],
+                                            test_environment.external_config["ports"]["http"][0])
             repeat = kwargs["repeat"]
             for repeat_count in xrange(repeat):
                 if repeat > 1:
                     logger.info("Repetition %i / %i" % (repeat_count + 1, repeat))
 
-                test_ids, test_queues = test_loader.queue_tests(kwargs["test_types"],
-                                                                kwargs["chunk_type"],
-                                                                kwargs["total_chunks"],
-                                                                kwargs["this_chunk"])
+
                 unexpected_count = 0
-                logger.suite_start(test_ids, run_info)
+                logger.suite_start(test_loader.test_ids, run_info)
                 for test_type in kwargs["test_types"]:
                     logger.info("Running %s tests" % test_type)
-                    tests_queue = test_queues[test_type]
+
+                    for test in test_loader.disabled_tests[test_type]:
+                        logger.test_start(test.id)
+                        logger.test_end(test.id, status="SKIP")
 
                     executor_cls = executor_classes.get(test_type)
                     executor_kwargs = get_executor_kwargs(base_server,
@@ -604,20 +697,22 @@ def run_tests(config, tests_root, metadata_root, product, **kwargs):
                                      (test_type, product))
                         continue
 
+
                     with ManagerGroup("web-platform-tests",
                                       kwargs["processes"],
+                                      test_source_cls,
+                                      test_source_kwargs,
                                       browser_cls,
                                       browser_kwargs,
                                       executor_cls,
                                       executor_kwargs,
                                       kwargs["pause_on_unexpected"]) as manager_group:
                         try:
-                            manager_group.start(tests_queue)
+                            manager_group.run(test_type, test_loader.tests)
                         except KeyboardInterrupt:
                             logger.critical("Main thread got signal")
                             manager_group.stop()
                             raise
-                        manager_group.wait()
                     unexpected_count += manager_group.unexpected_count()
 
                 unexpected_total += unexpected_count
